from datetime import datetime
from typing import Literal, Optional, TypedDict

from typing_extensions import NotRequired

from billing.types._billing_entity import BillingEntityWithTimestamps


class Agreement(BillingEntityWithTimestamps):
    name: str
    type: Literal["generic", "named", "additional"]
    offer_id: str
    signed_at: Optional[datetime]


class AgreementWithTerms(Agreement):
    terms: str


class AgreementListParams(TypedDict):
    customer_id: NotRequired[str]
    order_id: NotRequired[str]
    offer_id: NotRequired[str]
    is_signed: NotRequired[bool]



from datetime import datetime
from typing import Optional, TypedDict

from typing_extensions import NotRequired

from billing.types._billing_entity import BillingEntityWithTimestamps
from billing.types._common import CancellationFeedback


class Offer(BillingEntityWithTimestamps):
    order_id: str
    cancel_at: Optional[datetime]
    cancellation_requested_at: Optional[datetime]
    cancellation_feedback: Optional[CancellationFeedback]
    cancellation_comment: Optional[str]


class OfferCancelPayload(TypedDict):
    cancel_at: NotRequired[datetime]
    cancellation_feedback: NotRequired[CancellationFeedback]
    cancellation_comment: NotRequired[str]



from typing import Literal

HTTPMethod = Literal["GET", "OPTIONS", "HEAD", "POST", "PUT", "PATCH", "DELETE"]

PeriodType = Literal["day", "week", "month", "year"]

CancellationFeedback = Literal[
    "customer_service",
    "low_quality",
    "missing_features",
    "switched_service",
    "too_complex",
    "too_expensive",
    "unused",
    "other",
]



import uuid
from typing import List

from typing_extensions import Unpack

from billing.services._billing_api_service import BillingAPIService
from billing.types import FeatureRecordPayload, FeatureUsageEvent, FeatureUsageSummary


class FeatureAPIService(BillingAPIService):
    def retrieve_usage_summary(
        self,
        codename: str,
        customer_auth_service_id: uuid.UUID,
    ) -> FeatureUsageSummary:
        return self._request(
            "GET",
            f"/v1/features/{codename}/usage/summary/",
            params={"customer_auth_service_id": customer_auth_service_id},
            data_model=FeatureUsageSummary,
        )

    async def retrieve_usage_summary_async(
        self,
        codename: str,
        customer_auth_service_id: uuid.UUID,
    ) -> FeatureUsageSummary:
        return await self._request_async(
            "GET",
            f"/v1/features/{codename}/usage/summary/",
            params={"customer_auth_service_id": customer_auth_service_id},
            data_model=FeatureUsageSummary,
        )

    def list_usage_summary(
        self,
        customer_auth_service_id: uuid.UUID,
        page_number: int = 1,
        page_size: int = 50,
    ) -> List[FeatureUsageSummary]:
        return self._request(
            "GET",
            "/v1/features/usage/summary/",
            params={
                "page": page_number,
                "page_size": page_size,
                "customer_auth_service_id": customer_auth_service_id,
            },
            data_model=FeatureUsageSummary,
            batch_mode=True,
        )

    async def list_usage_summary_async(
        self,
        customer_auth_service_id: uuid.UUID,
        page_number: int = 1,
        page_size: int = 50,
    ) -> List[FeatureUsageSummary]:
        return await self._request_async(
            "GET",
            "/v1/features/usage/summary/",
            params={
                "page": page_number,
                "page_size": page_size,
                "customer_auth_service_id": customer_auth_service_id,
            },
            data_model=FeatureUsageSummary,
            batch_mode=True,
        )

    def record_usage(
        self,
        codename: str,
        **payload: Unpack[FeatureRecordPayload],
    ) -> FeatureUsageEvent:
        return self._request(
            "POST",
            f"/v1/features/{codename}/usage/record/",
            json=payload,
            data_model=FeatureUsageEvent,
        )

    async def record_usage_async(
        self,
        codename: str,
        **payload: Unpack[FeatureRecordPayload],
    ) -> FeatureUsageEvent:
        return await self._request_async(
            "POST",
            f"/v1/features/{codename}/usage/record/",
            json=payload,
            data_model=FeatureUsageEvent,
        )

    def refund_usage(self, feature_usage_event_id: str) -> FeatureUsageEvent:
        return self._request(
            "POST",
            f"/v1/features/usage/{feature_usage_event_id}/refund/",
            data_model=FeatureUsageEvent,
        )

    async def refund_usage_async(self, feature_usage_event_id: str) -> FeatureUsageEvent:
        return await self._request_async(
            "POST",
            f"/v1/features/usage/{feature_usage_event_id}/refund/",
            data_model=FeatureUsageEvent,
        )



import hashlib
import hmac
import time
from typing import Optional, Union

from billing._constants import TERMINAL_WEBHOOK_SECRET_KEY_PREFIX
from billing._exceptions import SignatureVerificationError
from billing._utils import json_loads
from billing.types import WebhookEvent


class Webhook:
    DEFAULT_TOLERANCE_SECONDS = 300

    @classmethod
    def construct_event(
        cls,
        payload: Union[bytes, str],
        signature: str,
        signature_timestamp: Union[str, int],
        terminal_webhook_secret_key: str,
        tolerance_seconds: Optional[int] = DEFAULT_TOLERANCE_SECONDS,
    ) -> WebhookEvent:
        if not terminal_webhook_secret_key.startswith(TERMINAL_WEBHOOK_SECRET_KEY_PREFIX):
            raise ValueError(
                f"Invalid terminal webhook secret key. Must starts with `{TERMINAL_WEBHOOK_SECRET_KEY_PREFIX}`."
            )

        if isinstance(payload, bytes):
            payload = payload.decode("utf-8")

        cls.verify_signature(
            payload=payload,
            signature=signature,
            signature_timestamp=signature_timestamp,
            terminal_webhook_secret_key=terminal_webhook_secret_key,
            tolerance_seconds=tolerance_seconds,
        )

        return WebhookEvent.parse(json_loads(payload))

    @classmethod
    def verify_signature(
        cls,
        payload: str,
        signature: str,
        signature_timestamp: Union[str, int],
        terminal_webhook_secret_key: str,
        tolerance_seconds: Optional[int] = None,
    ) -> None:
        if tolerance_seconds and int(signature_timestamp) < time.time() - tolerance_seconds:
            raise SignatureVerificationError(f"Timestamp ({signature_timestamp}) outside of the tolerance zone.")

        excepted_signature = cls._compute_excepted_signature(
            payload=payload,
            signature_timestamp=signature_timestamp,
            terminal_webhook_secret_key=terminal_webhook_secret_key,
        )

        if not hmac.compare_digest(signature, excepted_signature):
            raise SignatureVerificationError("Signature verification failed for given payload.")

    @staticmethod
    def _compute_excepted_signature(
        payload: str,
        signature_timestamp: Union[str, int],
        terminal_webhook_secret_key: str,
    ) -> str:
        secret = f"{signature_timestamp}.{payload}".encode()
        mac = hmac.new(
            terminal_webhook_secret_key.encode("utf-8"),
            msg=secret,
            digestmod=hashlib.sha256,
        )
        return mac.hexdigest()



from typing import List

from billing.types._invoice import InvoiceWithItemsBundle
from billing.types._offer import Offer
from billing.types._offer_product_plan import OfferProductPlan
from billing.types._order import Order
from billing.types._product import Product
from billing.types._product_plan import ProductPlan


class OfferPlansBundle(Offer):
    offer_product_plans: List[OfferProductPlan]


class OfferInvoicesBundle(Offer):
    invoices: List[InvoiceWithItemsBundle]


class OrderOffersBundle(Order):
    offers: List[OfferPlansBundle]


class OrderWithOffersAndInvoicesBundle(Order):
    offers: List[OfferInvoicesBundle]


class ProductPlanWithImagesBundle(Product):
    product_plans: List[ProductPlan]



from typing import Any, Callable

loads: Callable[[str], Any]
try:
    import orjson

    loads = orjson.loads
except ImportError:
    try:
        import ujson

        loads = ujson.loads
    except ImportError:
        import json

        loads = json.loads


def json_loads(text: str) -> Any:
    return loads(text)



from datetime import datetime
from typing import Any, Dict

from billing.types._billing_entity import BillingObject


class WebhookEvent(BillingObject):
    id: str
    event_type: str
    attempt_number: int
    published_at: datetime
    object: Dict[str, Any]



from billing._billing_client import BillingClient as BillingClient
from billing._exceptions import AuthenticationError as AuthenticationError
from billing._exceptions import BillingAPIError as BillingAPIError
from billing._exceptions import BillingError as BillingError
from billing._exceptions import FeatureUsageLimitExceededError as FeatureUsageLimitExceededError
from billing._exceptions import RateLimitError as RateLimitError
from billing._exceptions import SignatureVerificationError as SignatureVerificationError
from billing._webhook import Webhook as Webhook
from billing.types import Agreement as Agreement
from billing.types import AgreementListParams as AgreementListParams
from billing.types import AgreementWithTerms as AgreementWithTerms
from billing.types import BillingEntity as BillingEntity
from billing.types import BillingEntityWithTimestamps as BillingEntityWithTimestamps
from billing.types import BillingObject as BillingObject
from billing.types import FeatureRecordPayload as FeatureRecordPayload
from billing.types import FeatureUsageEvent as FeatureUsageEvent
from billing.types import FeatureUsageSummary as FeatureUsageSummary
from billing.types import HTTPXPayload as HTTPXPayload
from billing.types import Image as Image
from billing.types import Invoice as Invoice
from billing.types import InvoiceListParams as InvoiceListParams
from billing.types import InvoiceWithItemsBundle as InvoiceWithItemsBundle
from billing.types import Offer as Offer
from billing.types import OfferCancelPayload as OfferCancelPayload
from billing.types import OfferInvoicesBundle as OfferInvoicesBundle
from billing.types import OfferPlansBundle as OfferPlansBundle
from billing.types import OfferProductPlan as OfferProductPlan
from billing.types import OfferProductPlanWithProduct as OfferProductPlanWithProduct
from billing.types import Order as Order
from billing.types import OrderCreatePayload as OrderCreatePayload
from billing.types import OrderListParams as OrderListParams
from billing.types import OrderOffersBundle as OrderOffersBundle
from billing.types import OrderWithOffersAndInvoicesBundle as OrderWithOffersAndInvoicesBundle
from billing.types import Product as Product
from billing.types import ProductPlan as ProductPlan
from billing.types import ProductPlanWithImagesBundle as ProductPlanWithImagesBundle
from billing.types import ProductPlanWithProduct as ProductPlanWithProduct

__all__ = (
    "BillingClient",
    "AuthenticationError",
    "BillingAPIError",
    "BillingError",
    "FeatureUsageLimitExceededError",
    "RateLimitError",
    "SignatureVerificationError",
    "Webhook",
    "Agreement",
    "AgreementListParams",
    "AgreementWithTerms",
    "BillingEntity",
    "BillingEntityWithTimestamps",
    "BillingObject",
    "FeatureUsageEvent",
    "FeatureRecordPayload",
    "FeatureUsageSummary",
    "HTTPXPayload",
    "Image",
    "Invoice",
    "InvoiceListParams",
    "InvoiceWithItemsBundle",
    "Offer",
    "OfferCancelPayload",
    "OfferInvoicesBundle",
    "OfferPlansBundle",
    "OfferProductPlan",
    "OfferProductPlanWithProduct",
    "OfferProductPlanWithProduct",
    "Order",
    "OrderCreatePayload",
    "OrderListParams",
    "OrderOffersBundle",
    "OrderWithOffersAndInvoicesBundle",
    "Product",
    "ProductPlan",
    "ProductPlanWithProduct",
    "ProductPlanWithImagesBundle",
)
